/*
 * tmos - beer.c
 * 	version 0.0.0
 * 
 * Copyright (C) 2009, 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/crc16.h>
#include <tmos/jiffies.h>
#include <tmos/mb-slave.h>
#include <tmos/watchdog.h>
#include <tmos/trt.h>
#include <tmos/port.h>
#include <tmos/net.h>
#include <tmos/adc.h>
		
#define AR3C_CHD_EOF		0x04
#define AR3C_CHD_DEVICE_IOCTL	0x48

#define AR3C_CHD_IOCTL_SET_BAUDRATE	0x20
#define AR3C_CHD_IOCTL_SET_FREQ		0x21

struct ar3c_chd_xioctl {
	u16 id;			/* device/master/conn id */
	u16 cmd;		/* command number */
	u8 arg_len;		/* argument length */
	u8 arg[0];		/* argument data */
} __attribute__ ((packed));

#define DEVICE_BUILTIN		0xffff

static void beer_init(void) __reentrant
{
	port_t p;
	uint8_t address;
	msg_ssize_t ret;

	/*
	 * read network address from /dev/beersw
	 */
	p = port_open("/dev/beersw", O_RDONLY);

	ret = port_read(p, &address, 1);
	if (ret == 1) {
		if (address == 0)
			address = 1;

		net_ioctl(NET_IFACE_SERIAL0, NET_IOCTL_SET_ADDR_BYTE, &address);
	}
}

#ifdef CONFIG_BEER_CHD
static char chd_verify_magic(port_t p, off_t pos)
{
	uint32_t x;
	msg_ssize_t ret;

	ret = port_pread(p, (void __kernel *) &x, 4, pos);
	if ((ret != 4) || (x != be32_to_cpu(0x41523343)))
		return 1;
	pos += 4;

	ret = port_pread(p, (void __kernel *) &x, 4, pos);
	if ((ret != 4) || (x != be32_to_cpu(0x43484420)))
		return 1;
	pos += 4;
	return 0;
}

static char device_ioctl(port_t p, off_t pos)
{
	struct ar3c_chd_xioctl a;
	uint32_t x;
	uint8_t x8;
	msg_ssize_t ret;

	ret = port_pread(p, (void __kernel *) &a, sizeof(a), pos);
	if (ret != sizeof(a))
		return -1;
	pos += sizeof(a);

	a.id = be16_to_cpu(a.id);
	if (a.id != DEVICE_BUILTIN)
		return -1;

	a.cmd = be16_to_cpu(a.cmd);
	if (a.arg_len != 4)
		return -1;

	ret = port_pread(p, (void __kernel *) &x, 4, pos);
	if (ret != 4)
		return -1;
	x = be32_to_cpu(x);

	switch (a.cmd) {
	case AR3C_CHD_IOCTL_SET_BAUDRATE:
		x8 = x;
		net_ioctl(NET_IFACE_SERIAL0, NET_IOCTL_SET_SPEED, &x8);
		return a.arg_len;
#ifdef CONFIG_ADC
	case AR3C_CHD_IOCTL_SET_FREQ:
		adc_set_freq(x);
		return a.arg_len;
#endif
	}

	return -1;
}

static char beer_config_init(off_t pos) __reentrant
{
	port_t p;
	uint8_t type;
	msg_ssize_t ret;

	/*
	 * read config from /dev/eedata
	 */
	p = port_open(CONFIG_BEER_CHD_DEVICE, O_RDONLY);
	if (p == (port_t) -1)
		return 1;

	if (chd_verify_magic(p, pos))
		return 1;

	pos += 12;

	for (;;) {
		ret = port_pread(p, (void __kernel *) &type, 1, pos);
		if (ret != 1)
			return 1;
		pos++;

		switch (type) {
		case AR3C_CHD_EOF:
			return 0;
		case AR3C_CHD_DEVICE_IOCTL:
			ret = device_ioctl(p, pos);
			if (ret < 0)
				return 1;

			pos += sizeof(struct ar3c_chd_xioctl) + ret;
			break;
		default:
			return 1;
		}
	}
}
#else
#define beer_config_init(pos)	(0)
#endif

void app_main(void) REENTRANT
{
	beer_init();
	if (beer_config_init(0))
		beer_config_init(0x800);

	for (;;) {
		watchdog_refresh();

		trt_poll();
	}
}
