/*
 * tmos - pommax2-boot.c
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <tmos/kernel.h>
#include <tmos/watchdog.h>
#include <tmos/errno.h>
#include <tmos/port.h>
#include <asm/system.h>
#include "p2.h"

#define P2_VERSION_LEN		0
#define  P2_VERSION_MASK	0xf0
#define  P2_LEN_MASK		0x0f
#define P2_PROTO		1
#define  P2_PROTO_NOP		0
#define  P2_PROTO_ANALOG	1
#define  P2_PROTO_CONFIG	2
#define  P2_PROTO_BOOT		3

#define P2_FLASH_CODE		0
#define P2_FLASH_DATA		1

struct p2_boot_err {
	u8 cmd;
	s8 err;
};

#define P2_BOOT_STAT	1
struct p2_boot_flash_stat {
	u8 cmd;
	u8 flash;
};

struct p2_boot_flash_stat {
	u16 page_flags;
	u16 npages;
};

#define P2_BOOT_READ4		2
struct p2_boot_read4 {
	s16 data0;
	s16 data1;
	s16 data2;
	s16 data3;
};

#define P2_BOOT_WRITE		3
struct p2_boot_write {
	s16 data0;
};

#define P2_BOOT_ERASE_PAGE	4
struct p2_boot_erase {
	u8 cmd;
	u8 flash;
	u16 page;
};

#define P2_BOOT_READ_PAGE	5
struct p2_boot_read {
	u16 page;
};

#define P2_BOOT_WRITE_PAGE	6
struct p2_boot_read {
	u8 cmd;
	u8 flash;
	u16 page;
};

#define P2_BOOT_EXEC		7

#define P2_MTU	16
static u8 msg_buffer[P2_MTU + 2];
#define buffer_payload ((void __kernel *) &msg_buffer[4])

static port_t rom_port;
static port_t data_port;

static char p2_boot_stat(unsigned char len)
{
	struct p2_boot_flash_stat __kernel *stat = buffer_payload;

	switch (msg_buffer[3]) {
	case P2_FLASH_CODE:
		stat->page_flags = cpu_to_be16(6);	/* 64 byte */
		stat->npages = cpu_to_be16(896);	/* 56 kB */
		break;
	case P2_FLASH_DATA:
		stat->page_flags = cpu_to_be16(2);	/* 4 byte */
		stat->npages = cpu_to_be16(1024);	/* 4 kB */
		break;
	default:
		return -ENOENT;
	}
	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 3;
	return 0;
}

static page_t buffer_page;

static char p2_boot_read4(unsigned char len)
{
	struct p2_boot_read4 __kernel *msg = buffer_payload;
	unsigned char offset = msg_buffer[3];
	const u8 __highmem *v = page_to_virt(buffer_page);

	v += offset << 2;

	msg->data0 = cpu_to_be16(v[0]);
	msg->data1 = cpu_to_be16(v[1]);
	msg->data2 = cpu_to_be16(v[2]);
	msg->data3 = cpu_to_be16(v[3]);

	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 5;
	return 0;
}

static char p2_boot_write(unsigned char len)
{
	struct p2_boot_write __kernel *msg = buffer_payload;
	unsigned char offset = msg_buffer[3];
	u8 __highmem *v = page_to_virt(buffer_page);

	if (len < 2)
		return -EINVAL;

	v += offset;

	v[0] = msg_buffer[5];
	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 1;
	return 0;
}

static char p2_boot_read_page(unsigned char len) __reentrant
{
	off_t off;
	u8 __highmem *v = page_to_virt(buffer_page);
	unsigned char count;
	unsigned char chunk;
	unsigned char i;
	msg_ssize_t ret;
	port_t port;
	char buf[16];

	switch (msg_buffer[3]) {
	case P2_FLASH_CODE:
		count = 64;
		port = rom_port;
		off = be16_to_cpu(*((const u16 *) &msg_buffer[4])) << 6;
		break;
	case P2_FLASH_DATA:
		port = data_port;
		count = 4;
		off = be16_to_cpu(*((const u16 *) &msg_buffer[4])) << 2;
		break;
	default:
		return -ENOENT;
	}

	do {
		chunk = count;
		if (chunk > sizeof(buf))
			chunk = sizeof(buf);

		ret = port_pread(port, buf, chunk, off);
		if (ret == 0)
			break;
		if (ret < 0)
			return ret;

		off += ret;
		count -= ret;

		for (i = 0; i < chunk; i++)
			*v++ = buf[i];
	} while (count);

	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 1;
	return 0;
}

static char p2_boot_write_page(unsigned char len) __reentrant
{
	off_t off;
	u8 __highmem *v = page_to_virt(buffer_page);
	unsigned char count;
	unsigned char chunk;
	unsigned char i;
	msg_ssize_t ret;
	port_t port;
	char buf[16];

	switch (msg_buffer[3]) {
	case P2_FLASH_CODE:
		count = 64;
		port = rom_port;
		off = be16_to_cpu(*((const u16 *) &msg_buffer[4])) << 6;
		break;
	case P2_FLASH_DATA:
		port = data_port;
		count = 4;
		off = be16_to_cpu(*((const u16 *) &msg_buffer[4])) << 2;
		break;
	default:
		return -ENOENT;
	}

	do {
		chunk = count;
		if (chunk > sizeof(buf))
			chunk = sizeof(buf);

		for (i = 0; i < chunk; i++)
			buf[i] = *v++;

		ret = port_pwrite(port, buf, chunk, off);
		if (ret == 0)
			break;
		if (ret < 0)
			return ret;

		off += ret;
		count -= ret;
	} while (count);

	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 1;
	return 0;
}

/*
 * execute program at specified address addr.
 */
static void kexec(u16 addr)
{
	void (*prog)(void);

	prog = (void * __force)(void __code * __force) addr;
	prog();
}

static char p2_boot_exec(unsigned char len)
{
	u16 addr;

	if (len < 2)
		return -EINVAL;

	addr = pack16(msg_buffer[5], msg_buffer[3]);

	/* acknowledge command */
	msg_buffer[3] = 0;
	msg_buffer[P2_VERSION_LEN] = 1;
	p2_xmit(msg_buffer);

	kexec(addr);
	return 0;
}

static char parse_msg(void) __reentrant
{
	unsigned char len;
	char ret;

	if ((msg_buffer[P2_VERSION_LEN] & P2_VERSION_MASK) != 0)
		return 1;
	len = msg_buffer[P2_VERSION_LEN] & P2_LEN_MASK;

	if (len < 1)
		return 1;

	if (msg_buffer[P2_PROTO] != P2_PROTO_BOOT)
		return 1;

	switch (msg_buffer[2]) {
	case P2_BOOT_STAT:
		ret = p2_boot_stat(len);
		break;
	case P2_BOOT_READ4:
		ret = p2_boot_read4(len);
		break;
	case P2_BOOT_WRITE:
		ret = p2_boot_write(len);
		break;
	case P2_BOOT_READ_PAGE:
		ret = p2_boot_read_page(len);
		break;
	case P2_BOOT_WRITE_PAGE:
		ret = p2_boot_write_page(len);
		break;
	case P2_BOOT_EXEC:
		ret = p2_boot_exec(len);
		break;
	default:
		ret = -ENOSYS;
		break;
	}

	if (ret < 0) {
		msg_buffer[P2_VERSION_LEN] = 1;
		msg_buffer[3] = ret;
	}
	return 0;
}

static char message_ok(void) __reentrant
{
	unsigned char len;

	if ((msg_buffer[P2_VERSION_LEN] & P2_VERSION_MASK) != 0)
		return 0;
	len = msg_buffer[P2_VERSION_LEN] & P2_LEN_MASK;
	if (len < 1)
		return 0;
	if (msg_buffer[P2_PROTO] != P2_PROTO_BOOT)
		return 0;
	return 1;
}

#define BOOT_TIMEOUT	1000	/* ms */
static char boot_timeout(void) __reentrant
{
	unsigned int timeout = BOOT_TIMEOUT;
	char ret = 0;

	do {
		watchdog_refresh();

		msg_buffer[P2_VERSION_LEN] = 1;
		msg_buffer[P2_PROTO] = P2_PROTO_ANALOG;
		msg_buffer[2] = 0;
		msg_buffer[3] = 0;

		ret = p2_xmit(msg_buffer);
		if (ret && message_ok())
			break;
		mdelay(1);
		timeout--;
	} while (timeout);
	return ret;
}

static void try_kexec(void) __reentrant
{
	u16 addr;

	addr = 0x0000;
	if (*((const u8 __code *) addr) == 0xff)
		return;

	kexec(0);
}

void app_main(void) __reentrant
{
	char ret;

	ret = boot_timeout();
	if (!ret)
		try_kexec();

	buffer_page = page_alloc();
	BUG_ON(buffer_page == PAGE_NULL);
	rom_port = port_open("/dev/code", O_RDWR);
	BUG_ON(rom_port == -1);
	data_port = port_open("/dev/eedata", O_RDWR);
	BUG_ON(data_port == -1);

	for (;;) {
		watchdog_refresh();

		if (!ret || parse_msg()) {
			/*
			 * If we received no message or parse_msg() failed
			 * we need to send something. The pommax3-v2-0 does
			 * not support P2_PROTO_NOP and to not currupt
			 * message to user we must use P2_PROTO_ANALOG.
			 */
			msg_buffer[P2_VERSION_LEN] = 1;
			msg_buffer[P2_PROTO] = P2_PROTO_ANALOG;
			msg_buffer[2] = 0;
			msg_buffer[3] = 0;
		}

		/*
		 * the p2_xmit returns 1 if message was received.
		 */
		ret = p2_xmit(msg_buffer);
	}
}
