/*
 * tmos - bimp.c
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * Tiny RT BIMP modules support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <tmos/kernel.h>
#include <tmos/trt.h>
#include <tmos/adc.h>
#include <tmos/jiffies.h>
#include <asm/io.h>

#ifdef CONFIG_TRT_BIMP_BACKSTORE
#include <asm/eedata.h>
#endif

#ifdef CONFIG_TRT_BIMP_VELOCITY
static int16_t bimp_vel;
static int32_t bimp_pos;

#ifdef CONFIG_TRT_BIMP_AR2
static int16_t bimp_vel2;
static int8_t e2;	/* for error feedback */

#define __bimp_velocity()	(bimp_vel2)
#else
#define __bimp_velocity()	(bimp_vel)
#endif

#define A_SHIFT CONFIG_TRT_BIMP_A_SHIFT
#define A (1 << A_SHIFT)
#define K CONFIG_TRT_BIMP_GAIN

/**
 * trt_bimp_irq - bimp periodic interrupt
 *
 * This function is called from system timer interrupt.
 */
void trt_bimp_irq(void) __using(TIMER_BANK) __reentrant
{
	static int16_t os;
	static int8_t e;	/* for error feedback */
	int16_t s;
	int16_t v = 0;

	/*
	 * read card 0
	 *
	 * 16-bit counter allows velocity up to 1.6 Mimp/s.
	 */
	s = ioread16_timer(4);

	v = s - os;
	bimp_pos += v;
	os = s;

	/*
	 * First order ARI with 1-st order error feedback
	 */
	v = K * v - bimp_vel;
	v += e;
	s = v;
	v = v + (A / 2);
	e = s - (v & ~(A - 1));
       	v = v >> A_SHIFT;
	bimp_vel += v;

#ifdef CONFIG_TRT_BIMP_AR2
	/*
	 * Second First order ARI with 1-st order error feedback
	 */
	v = bimp_vel - bimp_vel2;
	v += e2;
	s = v;
	v = v + (A / 2);
	e2 = s - (v & ~(A - 1));
       	v = v >> A_SHIFT;
	bimp_vel2 += v;
#endif
}
#endif

#ifdef CONFIG_TRT_BIMP_BACKSTORE
static __bit power_fail = 1;
static __highmem unsigned int high_timeout = 100;

static void trt_bimp_pfail(void) __reentrant
{
	int32_t pos;

	local_irq_disable();
	pos = bimp_pos;
	local_irq_enable();
	eedata_write_page(1023, &pos);
}
#endif

void trt_bimp_read(void) __reentrant
{
	int16_t hi = 0;
	int16_t lo = 0;

#ifdef CONFIG_TRT_BIMP_BACKSTORE
	lo = trt_read_reg(CONFIG_TRT_BIMP_POWER_REG);
	if (lo < CONFIG_TRT_BIMP_TRIGGER_LOW) {
		if (!power_fail) {
			trt_bimp_pfail();
			power_fail = 1;

			/* this should be quite high to prevent
			 * flash wear */
			high_timeout = 500;
		}
	}
	if (lo > CONFIG_TRT_BIMP_TRIGGER_HIGH) {
		if (power_fail) {
			if (high_timeout) {
				high_timeout--;
			} else {
				eedata_erase_page(1023);
				power_fail = 0;
			}
		}
	}
#endif

#ifdef CONFIG_TRT_BIMP_VELOCITY
	local_irq_disable();
	lo = ((int16_t __data *) &bimp_pos)[0];
	hi = ((int16_t __data *) &bimp_pos)[1];
	local_irq_enable();
#else
	lo = ioread16(4);
	hi = 0;
#endif

	trt_write_reg(hi, CONFIG_TRT_BIMP_INPUTBASE);
	trt_write_reg(lo, CONFIG_TRT_BIMP_INPUTBASE + 1);

#ifdef CONFIG_TRT_BIMP_VELOCITY
	local_irq_disable();
	hi = __bimp_velocity();
	local_irq_enable();

	trt_write_reg(hi, CONFIG_TRT_BIMP_INPUTBASE + 2);
#endif
}

#ifdef CONFIG_TRT_BIMP_BACKSTORE
static void bimp_init(void) __reentrant
{
	int32_t pos;

	eedata_read_page(1023, &pos);

	local_irq_disable();
	bimp_pos = pos;
	local_irq_enable();
}

late_initcall(bimp_init)
#endif
